/****************************************************************
 *								*
 * DC Loadflow program 	$Revision: 1.7 $                       *
 *								*
 * File: loadflow.c						*
 *								*
 * Author: W. B. Langdon
 *         from J.E. Hodgson (Fortran original by A. Ekwue)	*
 *								*
 * Started: 22 October 1994
 *								*
 ****************************************************************/

/* Modifications (in reverse order)
WBL 12 Sep 1995 Reheck lost FORTRAN-77 interface
WBL 27 Oct 1994 Ensure all arrays are initialised
WBL 24 Oct 1994 Add caclculating power imbalance
*/

#include <stdio.h>
#include <float.h>
#include "loadflow.h"

/* This function forms the susceptance matrix for the network,
 * using the line reactance data */
void FormSusceptanceMatrix (int nn, int nl,
			    int* NNLK, int* NNLJ, float* XL, BOOL* MAINT,
			    /* output */
			    float B[NN][NN] )
{
	int i, j, k;
	
	for (i=1; i<=nn; i++){
		for (j=1; j<=nl; j++) B[i][j] = 0.0;
	}
	
	for (i=1; i<=nl; i++) {
	   if (MAINT[i]==FALSE) {
		k = NNLK[i];
		j = NNLJ[i];
		B[k][j] = B[k][j]-(1.0/XL[i]);
		B[j][k] = B[k][j];
		B[k][k] = B[k][k]+(1.0/XL[i]);
		B[j][j] = B[j][j]+(1.0/XL[i]);
	}}
	return;
}

/* The susceptance matrix is now used to form the BBUS matrix.
 * Power at each bus is used to find the PBUS array */
void ConversionOfBAndP (int nn, float B[NN][NN], float* P, 
			/* outputs */
			float BBUS[NN][NN], float* PBUS )
{
	int j,k;
	
	for (k=2; k<=nn; k++) {
		PBUS[k-1] = P[k];
		for (j=2; j<=nn; j++) {
			BBUS[k-1][j-1] = B[k][j];
		}
	}
	return;
}

/* This function is the 'guts' of the program. The BBUS matrix
 * is factorised, and PBUS is manipulated so that voltage angle,
 * ANG, may be found subsequently. */
void Decldu (int n, float a[NN][NN], float x[NL])

/* Note that 'a' accepts the BBUS matrix when the function is
 * called, and 'x' accepts the PBUS vector. */
{
	int e, f, g, h, i, j, k, l, m;
	float low[NL*NL], indl[NL*NL], dia[NL], locl[NL], xx;
	float upe[NL*NL], indu[NL*NL], locu[NL];
	
	memset(low, 0,sizeof(low));
	memset(indl,0,sizeof(indl));
	memset(dia, 0,sizeof(dia));
	memset(locl,0,sizeof(locl));
	memset(upe, 0,sizeof(upe));
	memset(indu,0,sizeof(indu));
	memset(locu,0,sizeof(locu));

	l = 0;
	m = 1;

/* Firstly, the BBUS matrix is factorised */
	for (i=1; i<=n; i++) {
		if (i != 1) {
		   for (j=1; j<=i-1; j++) {
			if (a[i][j] != 0.0) {
		   	     l = l + 1;
			     low[l] = a[i][j];
			     indl[l] = j;
			     for (k=j+1; k<=n; k++) {
			     a[i][k]=(a[i][k]-(a[i][j]*a[j][k]));
			    }
			}	
		   }
			locl[i]=l;
		}
		xx= (a[i][i]==0)? FLT_MAX : (1.0/a[i][i]);
/* 'dia' is calculated here. It is required for the manipulation
 * of PBUS */
		dia[i]=xx;
		locu[i]=m;
		if (i != n) {
			for (j=i+1; j<=n; j++) {
				if (a[i][j] != 0.0) {
					a[i][j]=(xx*a[i][j]);
					upe[m]=a[i][j];
					indu[m]=j;
					m=m+1;
				}
			}
		}
	}
/* Now PBUS (or x here) is manipulated. */
	l=1;
	x[1]=(x[1]*dia[1]);
	for (i=2; i<=n; i++) {
		m=locl[i];
		for (j=l; j<=m; j++) {
			l=l+1;
			e=indl[j];
			x[i]=x[i]-(x[e]*low[j]);
		}
		x[i]=(x[i]*dia[i]);
	}
	m=locu[n-1];
	for (h=1; h<=n-1; h++) {
		i=n-h;
		l=locu[i];
		g=m;
		for (j=l; j<=g; j++) {
			m=m-1;
			f=indu[j];
			x[i]=x[i]-(x[f]*upe[j]);
		}
	}

	return;
}

/* Now the BBUS matrix is factorised by calling the Decldu 
 * function. The angle at each busbar in rads is also calculated */
void FactorizeBBUSMatrix (int nn, float BBUS[NN][NN], float* PBUS, 
			  /* output */
			  float* ANG )
{	
	int i;

	Decldu (nn-1,BBUS,PBUS);
	ANG[1] = 0.0;
	for (i=1; i<=nn-1; i++) {
		ANG[i+1] = PBUS[i];
	}
	return;
}

/* Results are calculated, i.e. line MW flow, line loading (%),
 * and voltage angle. Note that the voltage angle is converted to
 * degrees before display */
void Compute(int nn, int nl,
	     int* NNLK, int* NNLJ, float* XL, BOOL* MAINT, float*P, float*ANG, 
	     /* output */
	     float* FLOW, float* IMBALANCE )
{
	int i, j, k;
	
	for (i=1; i<=nl; i++) {
	   k = NNLK[i];
	   j = NNLJ[i];
	   FLOW[i] = (MAINT[i]) ? 0 : (((1.0/XL[i])*(ANG[k]-ANG[j]))*100.0);
        }
	for (i=1; i<=nn; i++) {
	   IMBALANCE[i] = P[i]*100.0;
	   for (j=1; j<=nl; j++ ) {
		   if(NNLK[j]==i) IMBALANCE[i] -= FLOW[j];
		   if(NNLJ[j]==i) IMBALANCE[i] += FLOW[j];
	   }
        }
	return;
}



void Loadflow (int nn, int nl,
	       float* P, int* NNLK, int* NNLJ, float* XL, BOOL* MAINT, 
	       /*output*/
	       float* FLOW, float* IMBALANCE )
{
float B[NN][NN];
float PBUS[NN], BBUS[NN][NN];
float ANG[NN];

	FormSusceptanceMatrix (nn, nl, NNLK, NNLJ, XL, MAINT, B);
	ConversionOfBAndP (nn, B, P, BBUS, PBUS);
	FactorizeBBUSMatrix (nn, BBUS, PBUS, ANG );	
	Compute(nn, nl, NNLK, NNLJ, XL, MAINT, P, ANG, FLOW, IMBALANCE);
}


/*comments in dcload.for on NOD1,NOD2 are wring they are INTEGER*/

/*nb depends on -i2 or -r8 on SUN f77 command line
/*comments in dcload.for on LINEIN are wrong no size is specified*/
#define LOGICAL int

void loadflowf_ (int* nn_, int* nl_, 
		 double p[NN], int nnlk[NL], int nnlj[NL], 
		 double susc[NL], LOGICAL linein[NL], 
	       /*output*/
		 double flow[NL] )
{
	int nn; int nl;
	float P[NN]; float XL[NL];
	float FLOW[NL]; float IMBALANCE[NL];
	BOOL MAINT[NL];

	int i;
FILE* log;
FILE* dstrm;
dstrm = fopen("debug","w");
fprintf(dstrm,"loadflow_\n");
fprintf(dstrm,"nn_% d\n", *nn_);
fprintf(dstrm,"nl_% d\n", *nl_);

	nn = *nn_;
	nl = *nl_;

for(i=0;i<NL;i++) {
	fprintf(dstrm,"nnlk[%d] %d\n", i, nnlk[i]);
	fprintf(dstrm,"nnlj[%d] %d\n", i, nnlj[i]); }

for(i=0;i<NN;i++) {
	int x1,x2, X1;
	int* pt;

	fprintf(dstrm,"p[%d] %f\n", i, p[i]); 
	P[i+1] = p[i];

	pt=&p[i];
	x1=*pt++;
	x2=*pt;

	pt=&P[i+1];
	X1=*pt++;

	fprintf(dstrm,"x1 %x x2 %x X1 %x\n", x1, x2, X1);

	fprintf(dstrm,"P[%d] %f\n", i+1, P[i+1]); 
}

for(i=0;i<NL;i++) {
	fprintf(dstrm,"linein[%d] %d\n", i, linein[i]);
	MAINT[i+1] = !linein[i];
	fprintf(dstrm,"MAINT[%d] %d\n", i+1, MAINT[i+1]);
}

for(i=0;i<NL;i++) {
	int x1,x2;
	int* p;
	float f=0.406796;

	fprintf(dstrm,"susc[%d] %f\n", i, susc[i]);

	p=&susc[i];
	x1=*p++;
	x2=*p;

	fprintf(dstrm,"x1 %x x2 %x\n", x1, x2);

	XL[i+1] = 1.0/susc[i];

	p=&XL[i+1];
	x1=*p++;

	fprintf(dstrm,"x1 %x \n", x1);

	fprintf(dstrm,"XL[%d] %f\n", i+1, XL[i+1]);
}

/***********************************************************************
* Produce output file sw.log informat like that used in fn.cost and sw.cost
************************************************************************/

log = fopen("sw.log","w");
fprintf(log,"%d %d %d\n", nn, nl, 1);


for(i=1;i<=nn;i++) {
	fprintf(log,"%f ", P[i]/100.0); 
}

fprintf(log,"\n\n");

for(i=1;i<=nl;i++) {
	fprintf(log,"%d %d %f %f", nnlk[i-1], nnlj[i-1], XL[i], 9999.9);
	if(MAINT[i]) fprintf(log," M\n");
	else         fprintf(log," .\n");
}

/***********************************************************************
* Ok run C version of Loadflow now 
************************************************************************/

	Loadflow (nn, nl,
		  P, &nnlk[-1], &nnlj[-1], XL, MAINT, 
	       /*output*/
		  FLOW, IMBALANCE );

fprintf(dstrm,"Loadflow compledt\n");


for(i=0;i<NL;i++) {
	fprintf(dstrm,"FLOW[%d] %f\n", i+1, FLOW[i+1]);
/* Convert to FORTRAN offset and units FORTRAN code is expecting */
	flow[i] = FLOW[i+1]/100.0;
	fprintf(dstrm,"flow[%d] %f\n", i, flow[i]);
}

for(i=0;i<NN;i++) {
	fprintf(dstrm,"IMBALANCE[%d] %f\n", i+1, IMBALANCE[i+1]);
}

}


